@@ -4,35 +4,33 @@ use warnings;
 
 # Copyright 1999-2005 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Header: /var/cvsroot/gentoo-src/ufed/ufed.pl,v 1.28 2005/04/17 21:10:04 truedfx Exp $
+# $Header: /var/cvsroot/gentoo-src/ufed/ufed.pl,v 1.33 2005/05/01 10:00:47 truedfx Exp $
 
 use File::Temp qw(tempfile);
-use Switch;
 use Term::ReadKey;
 
-my $version = '0.37';
+my $version = '0.39';
 
 my %environment;
-$environment{$_}=[] for qw(USE); # INCREMENTALS, except we only need USE
+$environment{$_}={} for qw(USE); # INCREMENTALS, except we only need USE
 
-my @packages;
+my %packages;
 my @profiles;
-my @use_masked_flags;
-my @use_defaults_flags;
-my @make_defaults_flags;
-my @default_flags;
-my @make_conf_flags;
-my $make_conf_only;
-my @archs;
-my @all_flags;
+my %use_masked_flags;
+my %use_defaults_flags;
+my %make_defaults_flags;
+my %default_flags;
+my %make_conf_flags;
+my %archs;
+my %all_flags;
 my %use_descriptions;
 
 my @portagedirs;
 
-sub cleanup_env();
-sub finalise(@);
+sub finalise(%);
 sub flags_dialog();
 sub have_package($);
+sub merge(\%%);
 sub merge_env(\%);
 sub noncomments($);
 sub norm_path($$);
@@ -46,7 +44,6 @@ sub read_sh($);
 sub read_use_defaults();
 sub read_use_descs();
 sub read_use_mask();
-sub resolve_flags(@);
 sub save_flags(@);
 sub show_help();
 
@@ -56,82 +53,101 @@ read_use_mask;
 read_use_defaults;
 read_make_globals;
 read_make_defaults;
-#push @{$environment{USE}}, @use_defaults_flags;
 read_make_conf;
 read_archs;
-#merge_env %ENV;
-#cleanup_env;
 read_use_descs;
-@default_flags = resolve_flags @make_defaults_flags, @use_defaults_flags;
-@all_flags = resolve_flags @default_flags, @make_conf_flags;
-@use_masked_flags = resolve_flags @use_masked_flags, @archs;
 
-for(@make_defaults_flags, @use_defaults_flags, @make_conf_flags) {
-	next if $_ eq '-*';
-	my $flag = substr $_, /^-/;
-	$use_descriptions{$flag} = "(Unknown)"
-	if not exists $use_descriptions{$flag};
+%default_flags = %make_defaults_flags;
+merge %default_flags, %use_defaults_flags;
+
+%all_flags = %default_flags;
+merge %all_flags, %make_conf_flags;
+delete $all_flags{'*'};
+$all_flags{'-*'} = 1 if exists $make_conf_flags{'*'} && !$make_conf_flags{'*'};
+
+merge %use_masked_flags, %archs;
+
+for(keys %all_flags) {
+	$use_descriptions{$_} = "(Unknown)"
+	if not exists $use_descriptions{$_};
 }
+$use_descriptions{'-*'} = 'Never enable any flags other than those specified in /etc/make.conf';
 
-for(@use_masked_flags)
-{ delete $use_descriptions{$_} }
-
-my @flags;
+for(keys %use_masked_flags)
+{ delete $use_descriptions{$_} if $use_masked_flags{$_} }
 
+my %flags;
 DIALOG: {
 	my $rc = flags_dialog;
-	switch($rc) {
-		case -1 {
-			# error
-			print STDERR "fatal error: the dialog couldn't be opened\n";
-			exit 1
-		}
-		case 0 {
-			# save
-			# we don't check for use.masked flags here anymore
-			# the checks were broken. they were filtered out earlier anyway
-			if(-w '/etc/' || -w '/etc/make.conf' && -w '/etc/make.conf.old') {
-				@flags = finalise @flags;
-				save_flags @flags;
-			}
-			exit
-		}
-		case 1 {
-			# cancel
-			exit
-		}
-		case 2 {
-			# help
-			show_help;
-			redo DIALOG
+	if($rc==-1) {
+		# error
+		print STDERR "fatal error: the dialog couldn't be opened\n";
+		exit 1
+	} elsif($rc==0) {
+		# save
+		# we don't check for use.masked flags here anymore
+		# the checks were broken. they were filtered out earlier anyway
+		if(-w '/etc/' || -w '/etc/make.conf' && -w '/etc/make.conf.old') {
+			my @flags = finalise %flags;
+			save_flags @flags;
 		}
+		exit
+	} elsif($rc==1) {
+		# cancel
+		exit
+	} elsif($rc==2) {
+		# help
+		show_help;
+		redo DIALOG
 	}
 }
 
-sub cleanup_env() {
-	for(keys %environment) {
-		@{$environment{$_}} = sort(resolve_flags(@{$environment{$_}}))
-		if ref $environment{$_} eq 'ARRAY';
+sub finalise(%) {
+	my (%flags) = @_;
+	if(exists $flags{'-*'}) {
+		return sort keys %flags;
+	} else {
+		my(@enabled, @disabled);
+		my %all_flags;
+		@all_flags{keys %flags, keys %default_flags} = ();
+		for(sort keys %all_flags) {
+			next if $_ eq '*';
+			push @enabled,    $_  if $flags{$_} && !$default_flags{$_};
+			push @disabled, "-$_" if $default_flags{$_} && !$flags{$_};
+		}
+		return @enabled, @disabled;
 	}
 }
 
 sub have_package($) {
 	my ($cp) = @_;
-	return (grep { $cp eq $_ } @packages) > 0;
+	return $packages{$cp};
+}
+
+sub merge(\%%) {
+	my ($env, %env) = @_;
+	%{$env} = () if(exists $env{'*'});
+	$env->{$_} = $env{$_} for(keys %env);
 }
 
 sub merge_env(\%) {
 	my ($env) = @_;
 	for(keys %environment) {
-		if(ref $environment{$_} eq 'ARRAY') {
+		if(ref $environment{$_} eq 'HASH') {
 			if(exists $env->{$_}) {
-				$env->{$_} = [ split ' ', $env->{$_} ];
-				push @{$environment{$_}}, @{$env->{$_}};
+				my %split;
+				for(split ' ', $env->{$_}) {
+					my $off = s/^-//;
+					%split = () if($_ eq '*');
+					$split{$_} = !$off;
+				}
+				$env->{$_} = { %split };
+				merge %{$environment{$_}}, %{$env->{$_}};
 			}
 		}
 	}
 	for(keys %{$env}) {
-		if(ref $environment{$_} ne 'ARRAY') {
+		if(ref $environment{$_} ne 'HASH') {
 			$environment{$_} = $env->{$_};
 		}
 	}
@@ -140,12 +156,10 @@ sub merge_env(\%) {
 sub noncomments($) {
 	my ($fname) = @_;
 	my @result;
+	local $/;
 	if(open my $file, '<', $fname) {
-		while(<$file>) {
-			s/[ \t]*(?:#.*)?\n?$//;
-			next if $_ eq '';
-			push @result, $_;
-		}
+		@result = split /(?:[^\S\n]*(?:#.*)?\n)+/, <$file>."\n";
+		shift @result if @result>0 && $result[0] eq '';
 		close $file;
 	}
 	return @result;
@@ -153,7 +167,7 @@ sub noncomments($) {
 
 sub norm_path($$) {
 	my ($base, $path) = @_;
-	my @pathcomp = ($path =~ m!^/! || split(m!/!, $base), split(m!/!, $path));
+	my @pathcomp = ($path !~ m!^/! && split(m!/!, $base), split(m!/!, $path));
 	for(my $i=0;;$i++) {
 		last if $i == @pathcomp; # don't want to skip this with redo
 		if($pathcomp[$i] eq '' || $pathcomp[$i] eq '.') {
@@ -175,22 +189,22 @@ sub norm_path($$) {
 sub read_archs() {
 	for my $dir(@portagedirs) {
 		for(noncomments "$dir/profiles/arch.list") {
-			push @archs, $_;
+			$archs{$_} = 1;
 		}
 	}
 }
 
 sub read_make_conf() {
 	my %env = read_sh "/etc/make.conf";
-	push @make_conf_flags, @{$env{USE}} if exists $env{USE};
-	$make_conf_only = (grep { $_ eq '-*' } @make_conf_flags) > 0;
-	@portagedirs = $environment{PORTDIR}, split ' ', $environment{PORTDIR_OVERLAY};
+	merge %make_conf_flags, %{$env{USE}} if exists $env{USE};
+	@portagedirs = $environment{PORTDIR};
+	push @portagedirs, split ' ', $environment{PORTDIR_OVERLAY} if defined $environment{PORTDIR_OVERLAY};
 }
 
 sub read_make_defaults() {
 	for my $dir(@profiles) {
 		my %env = read_sh "$dir/make.defaults";
-		push @make_defaults_flags, @{$env{USE}} if exists $env{USE};
+		merge %make_defaults_flags, %{$env{USE}} if exists $env{USE};
 	}
 }
 
@@ -207,7 +221,7 @@ sub read_packages() {
 			if(open my $use, '<', "$_/USE") {
 				# could be shortened, but make sure not to strip off part of the name
 				s/-\d+(?:\.\d+)*\w?(?:_(?:alpha|beta|pre|rc|p)\d*)?(?:-r\d+)?$//;
-				push @packages, $_;
+				$packages{$_} = 1;
 				local $/;
 				my @provide = split ' ', <$provide>;
 				my @use = split ' ', <$use>;
@@ -216,10 +230,15 @@ sub read_packages() {
 					next if $pkg eq '(' || $pkg eq ')';
 					if($pkg !~ s/\?$//) {
 						$pkg =~ s/-\d+(?:\.\d+)*\w?(?:_(?:alpha|beta|pre|rc|p)\d*)?(?:-r\d+)?$//;
-						push @packages, $pkg;
+						$packages{$pkg} = 1;
 					} else {
 						my $musthave = $pkg !~ s/^!//;
-						if($musthave != (grep { $pkg eq $_ } @use) > 0) {
+						my $have = 0;
+						for(@use) {
+							if($pkg eq $_)
+							{ $have = 1; last }
+						}
+						if($musthave != $have) {
 							my $level = 0;
 							for($i++;$i<@provide;$i++) {
 								$level++ if $provide[$i] eq '(';
@@ -238,7 +257,7 @@ sub read_packages() {
 
 sub read_profiles() {
 	$_ = readlink '/etc/make.profile';
-	die "/etc/make.profile is not a symlink\n" if $!;
+	die "/etc/make.profile is not a symlink\n" if not defined $_;
 	@profiles = norm_path '/etc', $_;
 	PARENT: {
 		for(noncomments "$profiles[0]/parent") {
@@ -249,11 +268,12 @@ sub read_profiles() {
 }
 
 sub read_sh($) {
-	my $BLANK = qr{(?:[ \n\t]+|#.*)*};                        # whitespace and comments
-	my $IDENT = qr{([0-9A-Za-z~!\@#\$\%_*\\:;?,./\-+\{\}]+)}; # identifier; character list from portage_util.py
-	my $ASSIG = qr{=};                                        # assignment operator
-	my $UQVAL = qr{([0-9A-Za-z~!\@#\$\%_*\\:;?,./\-+\{\}]+)}; # unquoted value, same as ident
-	my $DQVAL = qr{"((?:[^\\"]|\\.)*)"}s;                     # doublequoted value
+	my $BLANK = qr{(?:[ \n\t]+|#.*)+};         # whitespace and comments
+	my $IDENT = qr{([^ \\\n\t'"{}=]+)};        # identifiers
+	my $ASSIG = qr{=};                         # assignment operator
+	my $UQVAL = qr{((?:[^ \\\n\t'"]+|\\.)+)}s; # unquoted value
+	my $SQVAL = qr{'([^']*)'};                 # singlequoted value
+	my $DQVAL = qr{"((?:[^\\"]|\\.)*)"}s;      # doublequoted value
 
 	my ($fname) = @_;
 	my %env;
@@ -261,27 +281,36 @@ sub read_sh($) {
 		{ local $/; $_ = <$file> }
 		eval {
 			for(;;) {
-				/\G$BLANK/gc or die;
-				last if pos == length; # don't use /\G$/gc because of multi zero length matching madness (see perlre)
+				/\G$BLANK/gc;
+				last if pos == length;
 				/\G$IDENT/gc or die;
 				my $name = $1;
-				/\G$BLANK/gc or die;
+				/\G$BLANK/gc;
 				/\G$ASSIG/gc or die;
-				/\G$BLANK/gc or die;
-				/\G$UQVAL/gc or
-				/\G$DQVAL/gc or die;
-				my $value = $1;
-				$value =~ s[
-					\\\n       | # backspace-newline
-					\\(.)      | # other escaped characters
-					(?:\$({)?)   # $
-					$IDENT       # followed by an identifier
-					(?(2)})      # optionally enclosed in braces
-				][
-					defined $3 ? $env{$3} || '' : # replace envvars
-					defined $1 ? $1             : # unescape escaped characters
-					             ''               # delete backspace-newlines
-				]gex;
+				/\G$BLANK/gc;
+				die if pos == length;
+				my $value = '';
+				for(;;) {
+					if(/\G$UQVAL/gc || /\G$DQVAL/gc) {
+						my $addvalue = $1;
+						$addvalue =~ s[
+							\\\n       | # backspace-newline
+							\\(.)      | # other escaped characters
+							\$({)?       # $
+							$IDENT       # followed by an identifier
+							(?(2)})      # optionally enclosed in braces
+						][
+							defined $3 ? $env{$3} || '' : # replace envvars
+							defined $1 ? $1             : # unescape escaped characters
+							             ''               # delete backspace-newlines
+						]gex;
+						$value .= $addvalue
+					} elsif(/\G$SQVAL/gc) {
+						$value .= $1
+					} else {
+						last
+					}
+				}
 				$env{$name} = $value;
 			}
 		};
@@ -297,7 +326,7 @@ sub read_use_defaults() {
 		for(noncomments "$dir/use.defaults") {
 			my ($flag, @packages) = split;
 			for(@packages)
-			{ push @use_defaults_flags, $flag if have_package $_ }
+			{ $use_defaults_flags{$flag} = 1 if have_package $_ }
 		}
 	}
 }
@@ -320,7 +349,8 @@ sub read_use_descs() {
 sub read_use_mask() {
 	for my $dir(@profiles) {
 		for(noncomments "$dir/use.mask") {
-			push @use_masked_flags, $_;
+			my $off = s/^-//;
+			$use_masked_flags{$_} = !$off;
 		}
 	}
 }
@@ -346,30 +376,13 @@ sub flags_dialog() {
 	for my $flag(sort { uc $a cmp uc $b } keys %use_descriptions) {
 		$items .= $flag . ' " ';
 
-		if(grep { $_ eq "-$flag" } @make_defaults_flags)
-		{ $items .= '(-' }
-		elsif(grep { $_ eq $flag } @make_defaults_flags)
-		{ $items .= '(+' }
-		else
-		{ $items .= '( ' }
+		$items .= exists $make_defaults_flags{$flag} ? $make_defaults_flags{$flag} ?'(+' :'(-' :'( ' ;
+		$items .= exists  $use_defaults_flags{$flag} ?  $use_defaults_flags{$flag} ? '+' : '-' : ' ' ;
+		$items .= exists     $make_conf_flags{$flag} ?     $make_conf_flags{$flag} ? '+)': '-)': ' )';
 
-		if(grep { $_ eq "-$flag" } @use_defaults_flags)
-		{ $items .= '-' }
-		elsif(grep { $_ eq $flag } @use_defaults_flags)
-		{ $items .= '+' }
-		else
-		{ $items .= ' ' }
+		$items .= ' ' . $use_descriptions{$flag} . '" ';
 
-		if(grep { $_ eq "-$flag" } @make_conf_flags)
-		{ $items .= '-) ' }
-		elsif(grep { $_ eq $flag } @make_conf_flags)
-		{ $items .= '+) ' }
-		else
-		{ $items .= ' ) ' }
-
-		$items .= $use_descriptions{$flag} . '" ';
-
-		if(grep { $_ eq $flag } @all_flags)
+		if($all_flags{$flag})
 		{ $items .= 'on' }
 		else
 		{ $items .= 'off' }
@@ -387,35 +400,30 @@ sub flags_dialog() {
 	           . '--item-help --help-button --help-status --checklist "Select desired set of USE '
 	           . 'flags from the list below:\\n(press SPACE to toggle, cursor keys to select)" '
 	           . $lines . ' ' . $cols . ' ' . ($lines - 8) . ' ' . $items;
-	switch(system('exec 3> '.$tempfile.' ; DIALOG_ESC="" dialog --file '.$cmdfile) >> 8) {
-		case [0] {
-			# OK
-			@flags = ();
-			open my $file, '<', $tempfile or die "Couldn't open temporary file\n";
-			while(<$file>) {
-				chomp;
-				push @flags, $_;
-			}
+	my $rc = system('exec 3> '.$tempfile.' ; DIALOG_ESC="" dialog --file '.$cmdfile) >> 8;
+	if($rc==0) {
+		# OK
+		open my $file, '<', $tempfile or die "Couldn't open temporary file\n";
+		local $/;
+		%flags = ();
+		$flags{$_} = 1 for split ' ', <$file>;
+		close $file;
+		return 0;
+	} elsif($rc==1||$rc==255) {
+		# Cancel, Esc
+		return 1;
+	} elsif($rc==2) {
+		# Help
+		if(open my $file, '<', $tempfile) {
+			<$file>; # skip past the first line which is just HELP and the flag description
+			local $/;
+			%all_flags = ();
+			$all_flags{$_} = 1 for split ' ', <$file>;
 			close $file;
-			return 0;
-		}
-		case [1, 255] {
-			# Cancel, Esc
-			return 1;
-		}
-		case [2] {
-			# Help
-			if(open my $file, '<', $tempfile) {
-				<$file>; # skip past the first line which is just HELP and the flag description
-				local $/;
-				@all_flags = split ' ', <$file>;
-				close $file;
-			}
-			return 2;
-		}
-		else {
-			return -1;
 		}
+		return 2;
+	} else {
+		return -1;
 	}
 }
 
@@ -516,45 +524,19 @@ Distributed under the terms of the GNU General Public License v2
 	     . $cols);
 }
 
-sub resolve_flags(@) {
-	my %results = ();
-	for(@_) {
-		if(/^-/) {
-			if($_ eq '-*') {
-				%results = ();
-			} else {
-				delete $results{substr $_, 1};
-			}
-		} else {
-			$results{$_} = 1;
-		}
-	}
-	return keys %results;
-}
-
-sub finalise(@) {
-	if($make_conf_only) {
-		return '-*', sort @_;
-	} else {
-		return
-			# all enabled flags that aren't in the default set
-			sort(              grep { my $flag = $_; !grep { $flag eq $_ } @default_flags } @_),
-			# all default flags that aren't in the enabled set, with a - in front
-			sort(map { "-$_" } grep { my $flag = $_; !grep { $flag eq $_ } @_ } @default_flags);
-	}
-}
-
 sub save_flags(@) {
-	my $UBLNK = qr{(?:                                        # blank, but scan for #USE=
-			[ \n\t]+ |                                        # whitespace
-			\#[ \t]*USE[ \t]*=.*(\n?) |                       # place a capture after #USE= (this is where the USE flags go)
-			\#.*)*}x;                                         # normal comments
-	my $BLANK = qr{(?:[ \n\t]+|#.*)*};                        # whitespace and comments
-	my $IDENT = qr{([0-9A-Za-z~!\@#\$\%_*\\:;?,./\-+\{\}]+)}; # identifier; character list from portage_util.py
-	my $ASSIG = qr{=};                                        # assignment operator
-	my $UQVAL = qr{[0-9A-Za-z~!\@#\$\%_*\\:;?,./\-+\{\}]+};   # unquoted value, same as ident
-	my $DQVAL = qr{"(?:[^\\"]|\\.)*"}s;                       # doublequoted value
-	my $BNDQV = qr{"(?:[^\\"]|\\.|\\\n())*"};                 # same as above, but scan for backslash-newline
+	my $BLANK = qr{(?:[ \n\t]+|#.*)+};              # whitespace and comments
+	my $UBLNK = qr{(?:                              # as above, but scan for #USE=
+		[ \n\t]+ |
+		\#[ \t]*USE[ \t]*=.*(\n?) | # place capture after USE=... line
+		\#.*)+}x;
+	my $IDENT = qr{([^ \\\n\t'"{}=]+)};             # identifiers
+	my $ASSIG = qr{=};                              # assignment operator
+	my $UQVAL = qr{(?:[^ \\\n\t'"]+|\\.)+}s;        # unquoted value
+	my $SQVAL = qr{'[^']*'};                        # singlequoted value
+	my $DQVAL = qr{"(?:[^\\"]|\\.)*"}s;             # doublequoted value
+	my $BNUQV = qr{(?:[^ \\\n\t'"]+|\\\n()|\\.)+}s; # unquoted value (scan for \\\n)
+	my $BNDQV = qr{"(?:[^\\"]|\\\n()|\\.)*"}s;      # doublequoted value (scan for \\\n)
 
 	my (@flags) = @_;
 	my $contents;
@@ -575,9 +557,10 @@ sub save_flags(@) {
 		my($ucs, $uce) = (length, length);
 		my $flags = '';
 		for(;;) {
-			/\G$UBLNK/gc or die;
+			if(/\G$UBLNK/gc) {
+				($ucs, $uce) = ($-[1], $+[1]) if defined $1;
+			}
 			last if pos == length;
-			($ucs, $uce) = ($-[1], $+[1]) if defined $1;
 			my $flagatstartofline = do {
 				my $linestart = 1+rindex $_, "\n", pos()-1;
 				my $line = substr($_, $linestart, pos()-$linestart);
@@ -585,19 +568,18 @@ sub save_flags(@) {
 			};
 			/\G$IDENT/gc or die;
 			my $name = $1;
-			/\G$BLANK/gc or die;
+			/\G$BLANK/gc;
 			/\G$ASSIG/gc or die;
-			/\G$BLANK/gc or die;
+			/\G$BLANK/gc;
+			die if pos == length;
 			if($name ne 'USE') {
-				/\G$UQVAL/gc or
-				/\G$DQVAL/gc or die;
+				/\G(?:$UQVAL|$SQVAL|$DQVAL)+/gc or die;
 			} else {
 				my $start = pos;
-				/\G$UQVAL/gc or
-				/\G$BNDQV/gc or die;
+				/\G(?:$BNUQV|$SQVAL|$BNDQV)+/gc or die;
 				my $end = pos;
 				# save whether user uses backslash-newline
-				my $bsnl = defined $1;
+				my $bsnl = defined $1 || defined $2;
 				# start of the line is one past the last newline; also handles first line
 				my $linestart = 1+rindex $_, "\n", $start-1;
 				# everything on the current line before the USE flags, plus one for the "
@@ -667,15 +649,14 @@ sub save_flags(@) {
 		} else { # if we replaced the flags, delete any further overrides
 			for(;;) {
 				my $start = pos;
-				/\G$BLANK/gc or die;
+				/\G$BLANK/gc;
 				last if pos == length;
 				/\G$IDENT/gc or die;
 				my $name = $1;
-				/\G$BLANK/gc or die;
+				/\G$BLANK/gc;
 				/\G$ASSIG/gc or die;
-				/\G$BLANK/gc or die;
-				/\G$UQVAL/gc or
-				/\G$DQVAL/gc or die;
+				/\G$BLANK/gc;
+				/\G(?:$UQVAL|$SQVAL|$DQVAL)+/gc or die;
 				my $end = pos;
 				if($name eq 'USE') {
 					substr($_, $start, $end-$start) = '';
